import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from .exceptions import UserNotificationException
from .subprocess import SubprocessExecutor, get_app_path  # nosec


@dataclass
class ScoopApp:
    name: str
    version: str
    path: Path
    manifest_file: Path
    bin_dirs: List[Path]


class ScoopWrapper:
    def __init__(self) -> None:
        self.scoop_path = self.get_scoop_path()

    @property
    def apps_directory(self) -> Path:
        return self.scoop_path.joinpath("apps")

    def install(self, scoop_file: Path) -> List[ScoopApp]:
        SubprocessExecutor([self.scoop_path, "install", scoop_file]).execute()
        # TODO: return the list of the installed apps
        return []

    def get_scoop_path(self) -> Path:
        scoop_path = get_app_path("scoop")
        if not scoop_path:
            raise UserNotificationException(
                "Scoop not found in PATH. Please run the build script again."
            )
        return scoop_path

    def parse_bin_dirs(
        self, bin_data: Union[str, List[Union[str, List[str]]]]
    ) -> List[Path]:
        """Parse the bin directory from the manifest file."""

        def get_parent_dir(bin_entry: Union[str, List[str]]) -> Optional[Path]:
            bin_path = (
                Path(bin_entry[0]) if isinstance(bin_entry, list) else Path(bin_entry)
            )
            return bin_path.parent if len(bin_path.parts) > 1 else None

        if isinstance(bin_data, str):
            return [parent for parent in [get_parent_dir(bin_data)] if parent]
        elif isinstance(bin_data, list):
            return [
                parent
                for parent in [get_parent_dir(bin_entry) for bin_entry in bin_data]
                if parent
            ]
        return []

    def get_installed_tools(self) -> List[ScoopApp]:
        installed_tools: List[ScoopApp] = []
        for manifest_file in self.apps_directory.glob("**/manifest.json"):
            app_directory: Path = manifest_file.parent
            # There is a directory level for the version of the tool
            tool_name: str = app_directory.parent.name
            with open(manifest_file) as f:
                manifest_data: Dict[str, Any] = json.load(f)
                tool_version: str = manifest_data.get("version", None)
                bin_dirs: List[Path] = self.parse_bin_dirs(manifest_data.get("bin", []))
                installed_tools.append(
                    ScoopApp(
                        name=tool_name,
                        version=tool_version,
                        path=app_directory,
                        manifest_file=manifest_file,
                        bin_dirs=bin_dirs,
                    )
                )
        return installed_tools



from pathlib import Path
from typing import List

from yanga.core.logger import logger
from yanga.ybuild.environment import BuildEnvironment
from yanga.ybuild.pipeline import Stage


class YangaInstall(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()

    def run(self) -> None:
        self.logger.info(
            f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}"
        )

    def get_dependencies(self) -> List[Path]:
        return []

    def get_outputs(self) -> List[Path]:
        return []



from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from pathlib import Path

from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter
from mashumaro import DataClassDictMixin

from yanga.core.cmd_line import Command
from yanga.core.exceptions import UserNotificationException
from yanga.core.logger import logger, time_it


@dataclass
class InitCommandConfig(DataClassDictMixin):
    project_dir: Path

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "InitCommandConfig":
        return cls.from_dict(vars(namespace))


class YangaInit:
    def __init__(self, config: InitCommandConfig) -> None:
        self.logger = logger.bind()
        self.config = config

    def run(self) -> None:
        self.logger.info(
            f"Run yanga init in '{self.config.project_dir.absolute().as_posix()}'"
        )
        self.create_project_from_template(self.config.project_dir)

    @staticmethod
    def create_project_from_template(output_dir: Path) -> None:
        # Create project from the cookiecutter project template
        this_dir = Path(__file__).parent
        project_dir_name = output_dir.name
        project_dir_parent = output_dir.parent
        try:
            cookiecutter(
                this_dir.joinpath("project-template").as_posix(),
                no_input=True,
                output_dir=project_dir_parent.as_posix(),
                extra_context={"project_dir_name": project_dir_name},
            )
        except OutputDirExistsException:
            raise UserNotificationException(
                f"Project directory '{output_dir}' is not empty. "
                "Remove it or use another name."
            )


class InitCommand(Command):
    def __init__(self) -> None:
        super().__init__("init", "Init a yanga project")
        self.logger = logger.bind()

    @time_it()
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        YangaInit(InitCommandConfig.from_namespace(args)).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--project-dir", help="Project directory", default=Path("."), type=Path
        )



import dataclasses
from abc import ABC, abstractmethod
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import Dict, List, Type

from .docs_utils import fulfills
from .logger import logger


class Command(ABC):
    def __init__(self, name: str, description: str) -> None:
        self.name = name
        self.description = description
        self.parser: ArgumentParser

    @abstractmethod
    def run(self, args: Namespace) -> int:
        """Run the command with the provided arguments."""

    def register_parser(self, parser_adder) -> None:  # type: ignore
        """Register the command parser.

        :param parser_adder: The return value of ``ArgumentParser.add_subparsers()``
        """
        self.parser = parser_adder.add_parser(
            self.name, help=self.description, exit_on_error=False
        )
        self._register_arguments(self.parser)

    @abstractmethod
    def _register_arguments(self, parser: ArgumentParser) -> None:
        """Register arguments for the command."""


class CommandLineHandler:
    """Handles the command line interface."""

    def __init__(self, commands: Dict[str, Command], parser: ArgumentParser) -> None:
        super().__init__()
        self.commands = commands
        self.parser = parser
        self.logger = logger.bind()

    @fulfills("REQ-CMDLINE_COMMAND_ARGS-0.0.1", "REQ-CMDLINE_COMMAND_EXEC-0.0.1")
    def run(self, args: List[str]) -> int:
        try:
            parsed_args = self.parser.parse_args(args)
        except ArgumentError as e:
            self.logger.error(f"Argument error: {e}")
            self.parser.print_help()
            return 1
        if (args is None) or (len(args) == 0):
            self.logger.debug("No command provided")
            self.parser.print_help()
            return 1
        command = self.commands.get(args[0], None)
        if command:
            return command.run(parsed_args)
        else:
            self.logger.error(f"Command {args[0]} not registered")
            return 1


class CommandLineHandlerBuilder:
    """Builds a command line handler."""

    def __init__(self) -> None:
        self.commands: Dict[str, Command] = {}
        self.parser = ArgumentParser(
            prog="yanga", description="Yanga CLI", exit_on_error=False
        )
        self.subparsers = self.parser.add_subparsers(title="Commands", dest="command")

    def create(self) -> CommandLineHandler:
        return CommandLineHandler(self.commands, self.parser)

    @fulfills("REQ-CMDLINE_REGISTER_COMMANDS-0.0.1", "REQ-CMDLINE_DUPLICATION-0.0.1")
    def add_command(self, command: Command) -> "CommandLineHandlerBuilder":
        """Add a command to the command line handler."""
        if self.commands.get(command.name, None) is not None:
            raise ValueError(f"Command {command.name} already exists")
        self.commands[command.name] = command
        command.register_parser(self.subparsers)
        return self


def register_arguments_for_config_dataclass(
    parser: ArgumentParser, config_dataclass: Type  # type: ignore
) -> None:
    """Helper function to register arguments for a dataclass.
    This avoid having to manually register arguments for each field of the dataclass."""
    if not dataclasses.is_dataclass(config_dataclass):
        raise TypeError(f"{config_dataclass.__name__} is not a dataclass.")

    for field_name, field in config_dataclass.__dataclass_fields__.items():
        parameter_default = (
            field.default if not field.default == dataclasses.MISSING else None
        )
        parameter_help = field.metadata.get(
            "help", f"Value for {field_name}. Default: {parameter_default}"
        )
        parameter_name = field_name.replace("_", "-")

        # TODO: this should be false if the type is Optional
        parameter_required = False if parameter_default else True

        parser.add_argument(
            f"--{parameter_name}",
            required=parameter_required,
            type=field.type,
            default=parameter_default,
            help=parameter_help,
        )



class UserNotificationException(Exception):
    """Exception raised to notify user about error detected.
    These are errors a user might be able to fix.
    This shall not be used to raise errors that are caused
    by the application missbehaviour.
    """

    pass



from pathlib import Path


class ProjectArtifactsLocator:
    """Provides paths to project artifacts."""

    config_file_name = "yanga.yaml"

    def __init__(self, project_root_dir: Path) -> None:
        self.project_root_dir = project_root_dir
        self.variants_dir = project_root_dir.joinpath("variants")
        self.platforms_dir = project_root_dir.joinpath("platforms")
        self.config_file = project_root_dir.joinpath(self.config_file_name)


class ProjectBuildArtifactsLocator(ProjectArtifactsLocator):
    def __init__(
        self,
        project_root_dir: Path,
        variant_name: str,
        build_config: str,
    ) -> None:
        super().__init__(project_root_dir)
        self.build_dir = project_root_dir / "build" / variant_name / build_config
        self.variant_dir = self.variants_dir / variant_name



import sys
from pathlib import Path
from sys import argv

from yanga.commands.build import BuildCommand
from yanga.commands.init import InitCommand
from yanga.core.cmd_line import CommandLineHandlerBuilder
from yanga.core.exceptions import UserNotificationException
from yanga.core.logger import logger, setup_logger, time_it


@time_it()
def do_run() -> None:
    logger.debug("Starting yanga")
    builder = CommandLineHandlerBuilder()
    builder.add_command(BuildCommand()).add_command(InitCommand())
    handler = builder.create()
    handler.run(argv[1:])


def main() -> int:
    try:
        setup_logger(Path(".yanga/logs/yanga.log"), clear=True)
        do_run()
    except UserNotificationException as e:
        logger.error(f"{e}")
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



from typing import Any, Callable


def validates(*req_ids: str) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def _doc(func: Callable[..., Any]) -> Callable[..., Any]:
        user_doc = func.__doc__ or ""
        req_ids_str = " ".join(req_ids)
        func.__doc__ = (
            user_doc
            + "\n"
            + f"""
         .. item:: UTEST-{func.__module__}.{func.__qualname__}
            :validates: {req_ids_str}
         """
        )
        return func

    return _doc


def fulfills(*req_ids: str) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def _doc(func: Callable[..., Any]) -> Callable[..., Any]:
        user_doc = func.__doc__ or ""
        req_ids_str = " ".join(req_ids)
        func.__doc__ = (
            user_doc
            + "\n"
            + f"""
         .. item:: IMPL-{func.__module__}.{func.__qualname__}
            :fulfills: {req_ids_str}
         """
        )
        return func

    return _doc



from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from typing import List, Type

from yanga.ybuild.config import PipelineConfig, StageConfig
from yanga.ybuild.environment import BuildEnvironment


class Stage(ABC):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        self.environment = environment
        self.output_dir = self.environment.artifacts_locator.build_dir / group_name

    @abstractmethod
    def run(self) -> None:
        """Run stage"""

    @abstractmethod
    def get_dependencies(self) -> List[Path]:
        """Get stage dependencies"""

    @abstractmethod
    def get_outputs(self) -> List[Path]:
        """Get stage outputs"""


@dataclass
class BuildStage:
    group_name: str
    _class: Type[Stage]


class PipelineLoader:
    def __init__(self, pipeline_config: PipelineConfig) -> None:
        self.pipeline_config = pipeline_config

    def load_stages(self) -> List[BuildStage]:
        result = []
        for group_name, stages_config in self.pipeline_config.items():
            result.extend(self._load_stages(group_name, stages_config))
        return result

    @staticmethod
    def _load_stages(
        group_name: str, stages_config: List[StageConfig]
    ) -> List[BuildStage]:
        result = []
        for stage_config in stages_config:
            stage_class_name = stage_config.class_name or stage_config.stage
            if not stage_config.file:
                # no file means that the stage is a built-in stage
                stage_class = PipelineLoader._load_builtin_stage(stage_class_name)
                result.append(BuildStage(group_name, stage_class))
            else:
                # TODO: load stage from file
                raise NotImplementedError("User defined stages are not supported yet")
        return result

    @staticmethod
    def _load_builtin_stage(stage_class_name: str) -> Type[Stage]:
        stage_module = import_module("yanga.ybuild.stages")
        stage_class = getattr(stage_module, stage_class_name)
        return stage_class


class StageRunner:
    """It checks if a stage must run in current environment.
    A stage shall run if any of the dependencies changed or one of the outputs is missing.
    All dependencies and outputs relevant information is stored in the stage output directory
    in a <stage>.deps file.
    """

    def __init__(self, environment: BuildEnvironment, stage: BuildStage) -> None:
        self.environment = environment
        self.stage = stage

    def run(self) -> None:
        stage = self.stage._class(self.environment, self.stage.group_name)
        if self._must_run(stage):
            stage.run()

    def _must_run(self, stage: Stage) -> bool:
        """A stage shall run if any of the dependencies changed or one of the outputs is missing.
        All dependencies and outputs relevant information is stored in the stage output directory
        in a <stage>.deps file."""
        return True



import json
import logging
import os
import re
import subprocess  # nosec
import sys
import venv
from pathlib import Path
from typing import List, Optional

logging.basicConfig(level=logging.DEBUG, filename="build.log")
logger = logging.getLogger("build")


this_dir = Path(__file__).parent
package_manager = "{{ cookiecutter.python_package_manager }}"


class UserNotificationException(Exception):
    pass


class SubprocessExecutor:
    def __init__(self, command: List[str | Path], cwd: Optional[Path] = None):
        self.command = " ".join([str(cmd) for cmd in command])
        self.current_working_directory = cwd

    def execute(self) -> None:
        result = None
        try:
            current_dir = (self.current_working_directory or Path.cwd()).as_posix()
            logger.info(f"Running command: {self.command} in {current_dir}")
            # print all virtual environment variables
            logger.debug(json.dumps(dict(os.environ), indent=4))
            result = subprocess.run(
                self.command,
                cwd=current_dir,
                capture_output=True,
                text=True,  # to get stdout and stderr as strings instead of bytes
            )  # nosec
            result.check_returncode()
        except subprocess.CalledProcessError as e:
            raise UserNotificationException(
                f"Command '{self.command}' failed with error:\n"
                f"{result.stderr if result else e}"
            )


class WindowsVirtualEnvironment:
    def __init__(self, venv_dir: Path) -> None:
        self.venv_dir = venv_dir
        self.activate_script = self.venv_dir.joinpath("Scripts/activate.bat")

    def create(self) -> None:
        """Create a virtual environment and install
        the python package manager with pip"""
        venv.create(self.venv_dir, with_pip=True)

    def pip(self, *args: str) -> None:
        """Run pip commands in the virtual environment"""
        pip_path = self.venv_dir.joinpath("Scripts/pip").as_posix()
        SubprocessExecutor([pip_path, *args], this_dir).execute()

    def run(self, *args: str) -> None:
        """Run commands in the virtual environment"""
        SubprocessExecutor(
            [f"cmd /c {self.activate_script.as_posix()} && ", *args], this_dir
        ).execute()


class Build:
    """TODO: use a VirtualEnvironment class to handle venv
    for different operating systems"""

    @property
    def venv_dir(self) -> Path:
        return this_dir / ".venv"

    @property
    def package_manager_name(self) -> str:
        match = re.match(r"^([a-zA-Z0-9_-]+)", package_manager)

        if match:
            return match.group(1)
        else:
            raise UserNotificationException(
                f"Could not extract the package manager name from {package_manager}"
            )

    def run(self) -> None:
        logger.info("Running project build script")
        virtual_env = WindowsVirtualEnvironment(self.venv_dir)
        virtual_env.create()
        virtual_env.pip("install", package_manager)
        virtual_env.run(self.package_manager_name, "install")
        # TODO: call yanga build
        # virtual_env.run("python -m yanga.ymain", "build")


def print_environment_info() -> None:
    str_bar = "".join(["-" for _ in range(80)])
    logger.debug(str_bar)
    logger.debug("Environment: \n" + json.dumps(dict(os.environ), indent=4))
    logger.info(str_bar)
    logger.info(f"Arguments: {sys.argv[1:]}")
    logger.info(str_bar)


def main() -> int:
    try:
        # print_environment_info()
        Build().run()
    except UserNotificationException as e:
        logger.error(e)
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())

if __name__ == "__test_main__":
    """This is used to execute the build script from a test and
    it shall not call sys.exit()"""
    main()



import shutil
import subprocess  # nosec
from pathlib import Path
from typing import List, Optional

from .exceptions import UserNotificationException
from .logger import logger


def get_app_path(app_name: str) -> Optional[Path]:
    """Return the path to the app if it is in the PATH, otherwise return None."""
    app_path = shutil.which(app_name)
    return Path(app_path) if app_path else None


class SubprocessExecutor:
    def __init__(self, command: List[str | Path], cwd: Optional[Path] = None):
        self.command = " ".join([str(cmd) for cmd in command])
        self.current_working_directory = cwd

    def execute(self) -> None:
        result = None
        try:
            logger.info(f"Running command: {self.command}")
            result = subprocess.run(
                self.command,
                cwd=(self.current_working_directory or Path.cwd()).as_posix(),
                capture_output=True,
                text=True,  # to get stdout and stderr as strings instead of bytes
            )  # nosec
            result.check_returncode()
        except subprocess.CalledProcessError as e:
            err_message = result.stderr if result else e
            raise UserNotificationException(
                f"Command '{self.command}' failed with error:\n{err_message}"
            )
